# __init__.py

__version__ = "0.1a1"

from .integrals import Integration
from .greedy import ReducedBasis, GramSchmidt
from .eim import EmpiricalMethods

__all__ = [Integration, ReducedBasis, EmpiricalMethods, GramSchmidt]



# - greedy.py -

# Copyright (c) 2020, Aarón Villanueva
# License: MIT
#   Full Text: https://gitlab.com/aaronuv/arby/-/blob/master/LICENSE

import numpy as np
from .integrals import Integration
from random import randint  # noqa: F401


# ==========================================
# Class for Iterated-Modified Gram-Schmidt
#      Orthonormalization Method
# ==========================================

def GS_add_element(h, basis, integration, a, max_iter):
    """
    Iterated modified Gram-Schmidt algorithm for building an orthonormal
    basis. Algorithm from Hoffman, `Iterative Algorithms for Gram-Schmidt
    Orthogonalization`. Given a function, h, find the corresponding basis
    function orthonormal to all previous ones.
    """
    norm = integration.norm(h)
    e = h / norm

    flag, ctr = 0, 1
    while flag == 0:
        for b in basis:
            e -= b * integration.dot(b, e)
        new_norm = integration.norm(e)
        if new_norm / norm <= a:
            norm = new_norm
            ctr += 1
            if ctr > max_iter:
                raise Exception("Gram-Schmidt: max number of iterations"
                                "reached.")
        else:
            flag = 1

    return [e / new_norm, new_norm]


def GramSchmidt(vectors, integration, a=0.5, max_iter=3):
    """Class for building an orthonormal basis using the iterated,
    modified Gram-Schmidt procedure.
    Input
    -----
    vectors: set of vectors to orthonormalize
    integration: instance of Integration class that defines the inner product

    Output is an array of orthonormal basis elements.
    """

    Nbasis, Nnodes = np.shape(vectors)
    functions = np.asarray(vectors)

    _, svds, _ = np.linalg.svd(functions)

    linear_indep_tol = 5e-15
    if min(svds) < linear_indep_tol:
        raise Exception("Functions are not linearly independent.")

    ortho_basis = []
    # First element of the basis is special, it's just normalized
    ortho_basis.append(integration.normalize(functions[0]))
    # For the rest of basis elements add them one by one by extending basis
    for new_basis_elem in functions[1:]:
        projected_element, _ = GS_add_element(
            new_basis_elem, ortho_basis, integration, a, max_iter
        )
        ortho_basis.append(projected_element)
    basis = np.array(ortho_basis)

    return basis


# ===============================
# Class for Reduced Basis Method
# ===============================

class ReducedBasis:
    """Class for standard reduced basis greedy algorithm."""

    def __init__(self, training_space, interval, rule="riemann"):
        self.training = np.array(training_space)
        self.Ntrain, self.Nsamples = training_space.shape
        self.loss = self.projection_error
        self.integration = Integration(interval=interval,
                                       num=self.Nsamples,
                                       rule=rule)

    def build_rb(self, index_seed=0, tol=1e-12, verbose=False):
        """Make a reduced basis using the standard greedy algorithm."""
        # In seed gives a null function, iterate to a new seed
        seed_function = self.training[index_seed]
        zero_function = np.zeros_like(seed_function)
        while np.allclose(seed_function, zero_function):
            index_seed = np.randint(1, self.Ntrain)
            seed_function = self.training[index_seed]

        # ====== Seed the greedy algorithm and allocate memory ================

        # Validate inputs
        assert self.Nsamples == np.size(
            self.integration.weights
        ), "Number of samples is inconsistent with quadrature rule."

        # Allocate memory for greedy algorithm arrays
        self.allocate(self.Ntrain, self.Nsamples, dtype=self.training.dtype)

        # Compute norms of the training space data
        self._norms = np.array([self.integration.norm(tt)
                               for tt in self.training])

        # Seed
        self.indices[0] = index_seed
        self.basis[0] = self.training[index_seed] / self._norms[index_seed]
        self.basisnorms[0] = self._norms[index_seed]
        self.proj_matrix[0] = self.integration.dot(self.basis[0],
                                                   self.training)

        # ===== Start greedy loop =============================================

        if verbose:
            print("\n Step", "\t", "Error")

        nn = 0
        sigma = 1.0
        while sigma > tol:
            nn += 1
            errs = self.loss(self.proj_matrix[:nn], norms=self._norms)
            next_index = np.argmax(errs)

            if next_index in self.indices:
                self.size = nn - 1
                self.trim(self.size)
                raise Exception("Index already selected: exiting"
                                "greedy algorithm.")

            self.indices[nn] = next_index
            self.errors[nn-1] = errs[next_index]
            self.basis[nn], self.basisnorms[nn] = GS_add_element(
                self.training[self.indices[nn]],
                self.basis[:nn],
                self.integration,
                a=0.5,
                max_iter=3,
                )
            self.proj_matrix[nn] = self.integration.dot(self.basis[nn],
                                                        self.training)
            sigma = errs[next_index]
            if verbose:
                print(nn, "\t", sigma)
        # Trim excess allocated entries
        self.size = nn
        self.trim(self.size)

    # ==== Auxiliary functions ================================================

    def allocate(self, Npoints, Nquads, dtype="complex"):
        """Allocate memory for numpy arrays used for making reduced basis"""
        self.errors = np.empty(Npoints, dtype="double")
        self.indices = np.empty(Npoints, dtype="int")
        self.basis = np.empty((Npoints, Nquads), dtype=dtype)
        self.basisnorms = np.empty(Npoints, dtype="double")
        self.proj_matrix = np.empty((Npoints, Npoints), dtype=dtype)

    def projection_error(self, proj_matrix, norms):
        """Square of the projection error of a function h on basis in terms
        of pre-computed projection matrix"""
        proj_norms = np.array([np.linalg.norm(proj_matrix[:, i])
                               for i in range(self.Ntrain)]
                              )
        return norms ** 2 - proj_norms**2

    def trim(self, num):
        """Trim arrays to have size num"""
        self.errors = self.errors[:num]
        self.indices = self.indices[:num]
        self.basis = self.basis[:num]
        self.proj_matrix = self.proj_matrix[:num]

    # ==== Validation functions ===============================================

    # This function inherites homology operation property from .dot method from
    # from integrals.py. Then h_vector may be an array of functions.
    def proj_error_from_basis(self, basis, h_vector):
        """Square of the projection error of a function h_vector on basis."""
        h_vector_sqnorm = self.integration.norm(h_vector).real
        inner_prod = np.array(
            [self.integration.dot(basis_elem, h_vector)
             for basis_elem in basis]
        )
        return h_vector_sqnorm**2 - np.linalg.norm(inner_prod)**2

    def project_on_basis(self, h, basis):
        """Project a function h onto the basis functions"""
        projected_function = 0.0
        for e in basis:
            projected_function += e * self.integration.dot(e, h)
        return projected_function



# --- integrals.py ---

# Copyright (c) 2020, Aarón Villanueva
# License: MIT
#   Full Text: https://gitlab.com/aaronuv/arby/-/blob/master/LICENSE


"""
Classes and functions for computing inner products of functions
"""

import numpy as np


#########################
# Some helper functions #
#########################


def _rate_to_num(a, b, rate):
    """Convert sample rate to sample numbers in [a,b]"""
    return np.floor(np.float(b - a) * rate) + 1


def _incr_to_num(a, b, incr):
    """Convert increment to sample numbers in [a,b]"""
    return _rate_to_num(a, b, 1.0 / incr)


def _make_rules(interval, rule_dict, num=None, rate=None, incr=None):
    """The workhorse for making quadrature rules"""

    # Validate inputs
    input_dict = {"num": num, "rate": rate, "incr": incr}

    assert type(interval) in [
        list,
        np.ndarray,
    ], "List or array input required."
    len_interval = len(interval)

    # Extract and validate the sampling method requested
    for kk, vv in input_dict.items():
        if vv is not None:
            key = kk
            value = input_dict[kk]
            if type(value) in [list, np.ndarray]:
                len_arg = len(value)
            else:
                len_arg = 1
                value = [value]
    assert (
        len_arg == len_interval - 1
    ), "Number of (sub)interval(s) does not equal number of arguments."

    # Generate nodes and weights for requested sampling
    nodes, weights = [], []
    for ii in range(len_arg):
        a, b = interval[ii: ii + 2]
        n, w = rule_dict[key](a, b, value[ii])
        nodes.append(n)
        weights.append(w)

    return [np.hstack(nodes), np.hstack(weights)]


def _nodes_weights(interval=None, num=None, rate=None, incr=None, rule=None):
    """Wrapper to make nodes and weights for integration classes"""

    # Validate inputs
    assert interval, "Input to `interval` must not be None."
    values = [num, rate, incr]
    if values.count(None) != 2:
        raise ValueError("Must give input for only one of num, rate, or incr.")
    if type(rule) is not str:
        raise TypeError("Input `rule` must be a string.")

    # Generate requested quadrature rule
    if rule in ["riemann"]:
        all_nodes, all_weights = QuadratureRules()[rule](
            interval, num=num, rate=rate, incr=incr
        )
    else:
        raise ValueError(f"Requested quadrature rule ({rule}) not available.")
    return all_nodes, all_weights, rule


##############################
# Class for quadrature rules #
##############################


class QuadratureRules:
    """Class for generating quadrature rules"""

    def __init__(self):
        self._dict = {"riemann": self.riemann}
        self.rules = list(self._dict.keys())

    def __getitem__(self, rule):
        return self._dict[rule]

    def riemann(self, interval, num=None, rate=None, incr=None):
        """
        Uniformly sampled array using Riemann quadrature rule
        over interval [a,b] with given sample number, sample rate
        or increment between samples.

        Input
        -----
        interval -- list indicating interval(s) for quadrature

        Options (specify only one)
        -------
        num  -- number(s) of quadrature points
        rate -- rate(s) at which points are sampled
        incr -- spacing(s) between samples

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """

        rule_dict = {"num": self._riemann_num}
        return _make_rules(interval, rule_dict, num=num, rate=rate, incr=incr)

    def _riemann_num(self, a, b, n):
        """
        Uniformly sampled array using Riemann quadrature rule
        over given interval with given number of samples

        Input
        -----
        a -- start of interval
        b -- end of interval
        n -- number of quadrature points

        Output
        ------
        nodes   -- quadrature nodes
        weights -- quadrature weights
        """
        nodes = np.linspace(a, b, num=n)
        weights = np.ones(n, dtype="double")
        weights[-1] = 0.0
        return [nodes, (b - a) / (n - 1.0) * weights]


class Integration:
    """Integrals for computing inner products and norms of functions"""

    def __init__(
        self,
        interval=None,
        num=None,
        rate=None,
        incr=None,
        rule="riemann",
        nodes=None,
        weights=None,
    ):

        self.nodes, self.weights, self.rule = _nodes_weights(
            interval, num, rate, incr, rule
        )

        self.integrals = ["integral", "dot", "norm", "normalize", "L2"]

    def integral(self, f):
        """Integral of a function"""
        return np.dot(self.weights, f)

    def dot(self, f, g):
        """Dot product between a function f and an array of functions g"""
        return np.dot(self.weights, (f.conjugate() * g).transpose())

    def norm(self, f):
        """Norm of function"""
        return np.sqrt(np.dot(self.weights, f.conjugate() * f).real)

    def normalize(self, f):
        """Normalize a function"""
        return f / self.norm(f)

    def L2(self, f):
        """L-2 norm"""
        return np.sqrt(np.dot(self.weights, f.conjugate() * f).real)



from setuptools import setup
import os

with open("README.md", "r") as f:
    long_description = f.read()

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
arby_init_path = os.path.join(BASE_DIR, "arby", "__init__.py")

with open(arby_init_path, "r") as f:
    for line in f:
        if line.startswith("__version__"):
            _, _, ARBY_VERSION = line.replace('"', "").split()
            break


setup(
    name="arby",
    version=ARBY_VERSION,
    # description="Short description",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Aarón Villanueva",
    author_email="aaron.villanueva@unc.edu.ar",
    url="https://gitlab.com/aaronuv/arby",
    packages=["arby"],
    install_requires=["numpy>=1.6", "scipy>=0.16"],
    test_suite="tests",
)



# --- eim.py ---

# Copyright (c) 2020, Aarón Villanueva
# License: MIT
#   Full Text: https://gitlab.com/aaronuv/arby/-/blob/master/LICENSE

import numpy as np
import copy


def two_norm(v):
    return np.linalg.norm(v, 2)


class EmpiricalMethods:
    """
    A class to select empirical nodes from a basis.

    """

    def __init__(self, B):
        self.B = B
        self.n = len(self.B)
        self.L = len(self.B[0])

    #################################
    # Construct the next Vandermonde matrix with eim nodes:
    # V(T0,...,Tn-1,Tn)

    def Vandermonde(self, van, nodes):
        v = copy.deepcopy(van)
        dim = len(v)
        new_node = nodes[dim]
        for i in range(dim):
            v[i].append(self.B[i][new_node])
        vertical = [self.B[dim][nodes[j]] for j in range(dim)]
        vertical.append(self.B[dim][new_node])
        v.append(vertical)
        return v

    ##########################################################
    # Function for EIM Classic #
    ##########################################################

    def eim(self, verbose=False):

        nodes = []
        lebesgue_norms = []
        k = []
        matrices = []

        new_node = np.argmax(np.abs(self.B[0]))
        if verbose:
            print(new_node)
        nodes.append(new_node)
        V = [[self.B[0][new_node]]]

        for i in range(1, self.n):
            matrices.append(V)
            base_in_nodes = [self.B[i][t] for t in nodes]
            invV = np.linalg.inv(V)
            leb = two_norm(invV)
            norm = two_norm(V)
            k.append(norm * leb)
            lebesgue_norms.append(leb)
            C = np.matmul(base_in_nodes, invV)
            base = self.B[:i]
            interpolant = np.matmul(C, base)
            residual = self.B[i] - interpolant
            new_node = np.argmax(abs(residual))
            if verbose:
                print(new_node)
            nodes.append(new_node)
            V = self.Vandermonde(V, nodes)
        matrices.append(V)
        invV = np.linalg.inv(V)
        leb = two_norm(invV)
        norm = two_norm(V)
        k.append(norm * leb)
        lebesgue_norms.append(leb)

        self.indices = nodes
        self.lebesgue = lebesgue_norms
        self.k = k
        self.matrices = matrices
